@@ -18,6 +18,10 @@
  *
  */
 
+
+/* _GNU_SOURCE is required to use struct ucred in glibc 2.8 */
+#define _GNU_SOURCE
+
 #include "unix-java.h"
 #include <stdio.h>
 #include <stdlib.h>
@@ -39,12 +43,8 @@
 extern "C" {
 #endif
 
-void handleerrno(JNIEnv *env)
+void throw(JNIEnv* env, int err, const char* msg)
 {
-   if (0 == errno) return;
-   int err = errno;
-   if (EAGAIN == err) return; // we read 0 bytes due to a timeout
-   const char* msg = strerror(err);
    jstring jmsg = (*env)->NewStringUTF(env, msg);
    jclass exc = (*env)->FindClass(env, "cx/ath/matthew/unix/UnixIOException");
    jmethodID cons = (*env)->GetMethodID(env, exc, "<init>", "(ILjava/lang/String;)V");
@@ -54,6 +54,15 @@ void handleerrno(JNIEnv *env)
    (*env)->Throw(env, exo);
    (*env)->DeleteLocalRef(env, exo);
 }
+
+void handleerrno(JNIEnv *env)
+{
+   if (0 == errno) return;
+   int err = errno;
+   if (EAGAIN == err) return; // we read 0 bytes due to a timeout
+   const char* msg = strerror(err);
+   throw(env, err, msg);
+}
    
 /*
  * Class:     cx_ath_matthew_unix_UnixServerSocket
@@ -194,8 +203,7 @@ JNIEXPORT jint JNICALL Java_cx_ath_matthew_unix_USInputStream_native_1recv
       tv.tv_usec = timeout;
       rv = select(sock+1, &rfds, NULL, NULL, &tv);
       rv = recv(sock, recvb, len, flags);
-      if (-1 == rv) { handleerrno(env); return -1; }
-      if (0 == rv) rv = -1;
+      if (-1 == rv) { handleerrno(env); rv = -1; }
       (*env)->ReleaseByteArrayElements(env, buf, cbuf, 0);
       return rv;
    } else  {
@@ -381,9 +389,9 @@ JNIEXPORT void JNICALL Java_cx_ath_matthew_unix_UnixSocket_native_1send_1creds
    cmsg->cmsg_len = CMSG_LEN(sizeof(struct ucred));
    /* Initialize the payload: */
    creds = (struct ucred *)CMSG_DATA(cmsg);
-   creds.pid = getpid();
-   creds.uid = getuid();
-   creds.gid = getgid();
+   creds->pid = getpid();
+   creds->uid = getuid();
+   creds->gid = getgid();
 #endif
 
    int rv = sendmsg(sock, &msg, 0);
@@ -399,7 +407,7 @@ JNIEXPORT jbyte JNICALL Java_cx_ath_matthew_unix_UnixSocket_native_1recv_1creds
   (JNIEnv *env, jobject o, jint sock, jintArray jcreds)
 {
    struct msghdr msg;
-   char buf = 0;
+   char iov_buf = 0;
    struct iovec iov;
    msg.msg_name = NULL;
    msg.msg_namelen = 0;
@@ -408,7 +416,7 @@ JNIEXPORT jbyte JNICALL Java_cx_ath_matthew_unix_UnixSocket_native_1recv_1creds
    msg.msg_iovlen = 1;
    msg.msg_control = NULL;
    msg.msg_controllen = 0;
-   iov.iov_base = &buf;
+   iov.iov_base = &iov_buf;
    iov.iov_len = 1;
 
 #ifdef SCM_CREDENTIALS
@@ -422,9 +430,9 @@ JNIEXPORT jbyte JNICALL Java_cx_ath_matthew_unix_UnixSocket_native_1recv_1creds
    recvmsg(sock, &msg, 0);
 
 #ifdef SCM_CREDENTIALS
-   for (cmsg = CMSG_FIRSTHDR(&msgh);
+   for (cmsg = CMSG_FIRSTHDR(&msg);
          cmsg != NULL;
-         cmsg = CMSG_NXTHDR(&msgh,cmsg)) {
+         cmsg = CMSG_NXTHDR(&msg,cmsg)) {
       if (cmsg->cmsg_level == SOL_SOCKET
             && cmsg->cmsg_type == SCM_CREDENTIALS) {
          creds = (struct ucred *) CMSG_DATA(cmsg);        
@@ -432,11 +440,15 @@ JNIEXPORT jbyte JNICALL Java_cx_ath_matthew_unix_UnixSocket_native_1recv_1creds
       }
    }
    if (NULL != creds) {
-      (*env)->SetIntArrayRegion(env, jcreds, 0, 3, creds);
+      jint cred_array[3];
+      cred_array[0] = creds->pid;
+      cred_array[1] = creds->uid;
+      cred_array[2] = creds->gid;
+      (*env)->SetIntArrayRegion(env, jcreds, 0, 3, &cred_array[0]);
    }
 #endif
 
-   return buf;
+   return iov_buf;
 }
 
 
